#!/usr/bin/env python

import pytest
import autoprop

def test_accessors():
    @autoprop   # (no fold)
    class Example(object):
        def get_attr(self):
            return 'attr'

    ex = Example()
    assert ex.attr == 'attr'

    @autoprop   # (no fold)
    class Example(object):
        def set_attr(self, attr):
            self._attr = 'new ' + attr

    ex = Example()
    ex.attr = 'attr'
    assert ex._attr == 'new attr'

    @autoprop   # (no fold)
    class Example(object):
        def del_attr(self):
            self._attr = None

    ex = Example()
    del ex.attr
    assert ex._attr is None

    @autoprop   # (no fold)
    class Example(object):
        def get_attr(self):
            return self._attr
        def set_attr(self, attr):
            self._attr = 'new ' + attr

    ex = Example()
    ex.attr = 'attr'
    assert ex.attr == 'new attr'

    @autoprop   # (no fold)
    class Example(object):
        def get_attr(self):
            return self._attr
        def set_attr(self, attr):
            self._attr = 'new ' + attr
        def del_attr(self):
            self._attr = None

    ex = Example()
    ex.attr = 'attr'
    assert ex.attr == 'new attr'
    del ex.attr
    assert ex.attr is None

def test_ignore_similar_names():
    @autoprop   # (no fold)
    class Example(object):
        def getattr(self):
            return 'attr'

    ex = Example()
    with pytest.raises(AttributeError):
        ex.attr

def test_ignore_empty_names():
    @autoprop   # (no fold)
    class Example(object):
        def get_(self):
            return 'get'

    ex = Example()
    with pytest.raises(AttributeError):
        getattr(ex, '')

def test_ignore_non_methods():
    @autoprop   # (no fold)
    class Example(object):
        get_attr = 'attr'

    ex = Example()
    with pytest.raises(AttributeError):
        ex.attr

def test_inheritance():
    @autoprop   # (no fold)
    class Parent(object):
        def get_attr(self):
            return 'parent'
        def get_overloaded_attr(self):
            return 'parent'

    @autoprop   # (no fold)
    class Child(Parent):
        def get_overloaded_attr(self):
            return 'child'

    parent = Parent()
    child = Child()

    assert parent.attr == 'parent'
    assert parent.overloaded_attr == 'parent'
    assert child.attr == 'parent'
    assert child.overloaded_attr == 'child'

def test_dont_overwrite_existing_attributes():
    @autoprop   # (no fold)
    class Example(object):
        attr = 'class var'
        def get_attr(self):
            return 'attr'

    ex = Example()
    assert ex.attr == 'class var'

def test_optional_arguments():
    @autoprop   # (no fold)
    class Example(object):
        def get_attr(self, pos=None, *args, **kwargs):
            return self._attr
        def set_attr(self, new_value, pos=None, *args, **kwargs):
            print('asdasd')
            self._attr = 'new ' + new_value
        def del_attr(self, pos=None, *args, **kwargs):
            self._attr = None

    ex = Example()
    ex.attr = 'attr'
    assert ex.attr == 'new attr'
    del ex.attr
    assert ex.attr == None

def test_getters_need_one_argument():
    @autoprop   # (no fold)
    class Example(object):
        def get_attr():
            return 'attr'

    ex = Example()
    with pytest.raises(AttributeError):
        ex.attr

    @autoprop   # (no fold)
    class Example(object):
        def get_attr(self, more_info):
            return 'attr'

    ex = Example()
    with pytest.raises(AttributeError):
        ex.attr

def test_setters_need_two_arguments():
    @autoprop   # (no fold)
    class Example(object):
        def set_attr(self):
            self._attr = 'no args'

    ex = Example()
    ex.attr = 'attr'
    with pytest.raises(AttributeError):
        assert ex._attr != 'no args'

    @autoprop   # (no fold)
    class Example(object):
        def set_attr(self, new_value, more_info):
            self._attr = 'two args'

    ex = Example()
    ex.attr = 'attr'
    with pytest.raises(AttributeError):
        assert ex._attr != 'two args'

def test_deleters_need_one_argument():
    @autoprop   # (no fold)
    class Example(object):
        def del_attr():
            pass

    ex = Example()
    with pytest.raises(AttributeError):
        del ex.attr

    @autoprop   # (no fold)
    class Example(object):
        def del_attr(self, more_info):
            pass

    ex = Example()
    with pytest.raises(AttributeError):
        del ex.attr


import sys
if sys.version_info[0] == 2:

    def test_complain_about_old_style_classes():
        with pytest.raises(TypeError) as err:
            @autoprop
            class Example:
                pass

        assert '@autoprop can only be used with new-style classes' in str(err)



#!/usr/bin/env python

def autoprop(cls):
    # These imports have to be inside autoprop(), otherwise the sys.modules 
    # hack below somehow makes them unavailable when the decorator is applied.
    import inspect, re
    from collections import defaultdict

    if not hasattr(cls, '__class__'):
        raise TypeError('@autoprop can only be used with new-style classes')

    accessors = defaultdict(dict)
    expected_num_args = {'get': 0, 'set': 1, 'del': 0}

    # The accessors we're searching for are considered methods in python2 and 
    # functions in python3.  Other than this, though, they behave the same.
    ismethod = lambda x: inspect.ismethod(x) or inspect.isfunction(x)

    for method_name, method in inspect.getmembers(cls, ismethod):
        accessor_match = re.match('(get|set|del)_(.+)', method_name)
        if not accessor_match:
            continue

        prefix, name = accessor_match.groups()
        arg_spec = inspect.getargspec(method)
        num_args = len(arg_spec.args) - len(arg_spec.defaults or ())
        num_args_minus_self = num_args - 1

        if num_args_minus_self != expected_num_args[prefix]:
            continue

        accessors[name][prefix] = method

    for name in accessors:
        if not name in cls.__dict__:
            setattr(cls, name, property(
                accessors[name].get('get'),
                accessors[name].get('set'),
                accessors[name].get('del'),
            ))

    return cls


autoprop.__version__ = '0.0.1'

# Abuse the import system so that the module itself can be used as a decorator.  
# This is a very simple intended only to cut-down on boilerplate, so I think 
# the trade-off between magicalness and ease-of-use is justified in this case.
import sys
sys.modules[__name__] = autoprop

